%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% computational action application
%
% Matthew Purver, 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library( lists ), [member/2, select/3] ).


% adjust/2 - adjust( +Tree1, ?Tree2 )
% Applies computational actions to Tree1, resulting in Tree2

% forcibly apply as many thinning / elimination rules as possible
% (this isn't really necessary, it's just to cut down on the number of partial trees in memory)
adjust( Tree1, Tree3 ) :-
	always( Tree1, Tree2 ),
	!,
	adjust( Tree2, Tree3 ).
% optionally apply any number of other rules
adjust( Tree1, Tree3 ) :-
	possible( Tree1, Tree2 ),
	adjust( Tree2, Tree3 ).
% or do nothing
adjust( Tree, Tree ) :-
	dbgnl.


% for efficiency, distinction between those rules that should always apply if they can,
% and those that aren't necessarily desirable
always_rules( [thinning, elimination, star-adjunction, conjunction] ).
possible_rules( [completion, introduction, prediction, merge,
		 link-adjunction, resolution] ).

always( Tree1, Tree2 ) :-
	always_rules( Rules ),
	member( Rule, Rules ),
	rule( Rule, Tree1, Tree2 ).

possible( Tree1, Tree2 ) :-
	possible_rules( Rules ),
	member( Rule, Rules ),
	rule( Rule, Tree1, Tree2 ).


% rule/3 - rule( ?RuleName, +Tree1, ?Tree2 )
% Defines the computational action RuleName in terms of input tree Tree1 and
% resulting output tree Tree2

% INTRODUCTION (English version)
rule( introduction,
      tree( Nodes, Pointer ), tree( [node( Pointer, [R0, R1 | Labels] )
				    | Nodes2], Pointer ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	member( ?ty(t), Labels ),
	dtr( 0, Pointer, D0 ),
	dtr( 1, Pointer, D1 ),
	R0 = ?([\/0],ty(e)),
	R1 = ?([\/1],ty(e>t)),
	\+ member( node( D0, _ ), Nodes ),
	\+ member( node( D1, _ ), Nodes ),
	\+ member( R0, Labels ),
	\+ member( R1, Labels ),
	dbg( 'comp: introduction at node ' ), dbg( Pointer ), dbgnl.

% % PREDICTION (general version)
% rule( prediction,
%       tree( Nodes, Pointer ), tree( [node( D, [?Req] ), node( Pointer, Labels2 )
% 				    | Nodes2], D ) ) :-
% 	select( node( Pointer, Labels ), Nodes, Nodes2 ),
% 	select( ?([\/N],Req), Labels, Labels2 ),
% 	dtr( N, Pointer, D ),
% 	\+ member( node( D, _ ), Nodes ),
% 	dbg( 'comp: prediction of ' ), dbg( D ), dbgnl.

% PREDICTION (English version for subject)
rule( prediction,
      tree( Nodes, Pointer ), tree( [node( D0, [?Req0] ),
				     node( Pointer, Labels2 ) | Nodes2], D0 ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	select( ?([\/0],Req0), Labels, Labels2 ),
	member( ?([\/1],_Req1), Labels ),
	dtr( 0, Pointer, D0 ),
	dtr( 1, Pointer, D1 ),
	\+ member( node( D0, _ ), Nodes ),
	\+ member( node( D1, _ ), Nodes ),
	dbg( 'comp: prediction of ' ), dbg( D0 ), dbgnl.

% PREDICTION (English version for predicate)
rule( prediction,
      tree( Nodes, Pointer ), tree( [node( D1, [?Req1] ),
				     node( Pointer, Labels2 ) | Nodes2], D1 ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	select( ?([\/1],Req1), Labels, Labels2 ),
	dtr( 0, Pointer, D0 ),
	dtr( 1, Pointer, D1 ),
	member( node( D0, _ ), Nodes ),
	\+ member( node( D1, _ ), Nodes ),
	dbg( 'comp: prediction of ' ), dbg( D1 ), dbgnl.

% *-ADJUNCTION
rule( star-adjunction,
      tree( Nodes, Pointer ), tree( [node( Node, [?ty( e ), ?fixed] ) | Nodes], Node ) ) :-
	member( node( Pointer, Labels ), Nodes ),
	member( ?ty( t ), Labels ),
	\+ member( ?([\/0],ty(e)), Labels ), % (i.e. no introduction has taken place)
	( Pointer = '0'; atom_concat( _, '2', Pointer ) ), % can't check singleton directly ...
	dtr( 0, Pointer, D0 ),
	dtr( 1, Pointer, D1 ),
	unfixed_dtr( Pointer, Node ),
	\+ member( node( D0, _ ), Nodes ),
	\+ member( node( D1, _ ), Nodes ),
	\+ member( node( Node, _ ), Nodes ),
	dbg( 'comp: *-adjunction of ' ), dbg( Node ), dbgnl.

% LINK-ADJUNCTION
rule( link-adjunction,
      tree( Nodes, Pointer ), tree( [node( LD, [?ty( t ), ?([\/ *],fo( Formula ))] ) | Nodes], LD ) ) :-
	member( node( Pointer, Labels ), Nodes ),
	member( ty( e ), Labels ),
	member( fo( Formula ), Labels ),
	linked_dtr( Pointer, LD ),
	\+ member( node( LD, _ ), Nodes ),
	dbg( 'comp: link-adjunction of ' ), dbg( LD ), dbgnl.


% THINNING
rule( thinning,
      tree( Nodes, Pointer ), tree( [node( Pointer, Labels2 ) | Nodes2], Pointer ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	select( ?Req, Labels, Labels2 ),
	check( tree( Nodes, Pointer ), Req ),
	dbg( 'comp: thin ' ), dbg( ?Req ), dbg( ' from ' ), dbg( Pointer ), dbgnl.

% THINNING
rule( thinning,
      tree( Nodes, Pointer ), tree( [node( Pointer, Labels2 ) | Nodes2], Pointer ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	select( ?(Dir,Req), Labels, Labels2 ),
	check( tree( Nodes, Pointer ), (Dir,Req) ),
	dbg( 'comp: thin ' ), dbg( ?(Dir,Req) ), dbg( ' from ' ), dbg( Pointer ), dbgnl.

% ELIMINATION
rule( elimination,
      tree( Nodes, Pointer ), tree( [node( Pointer, [ty( Type ), fo( Formula ) | Labels] ) | Nodes2], Pointer ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	\+ member( ty( _Type ), Labels ),
	\+ member( fo( _Formula ), Labels ),
	dtr( 0, Pointer, D0 ),
	dtr( 1, Pointer, D1 ),
	member( node( D0, Labels0 ), Nodes ),
	member( ty( Type0 ), Labels0 ),
	member( fo( Formula0 ), Labels0 ),
	member( node( D1, Labels1 ), Nodes ),
	member( ty( Type1 ), Labels1 ),
	member( fo( Formula1 ), Labels1 ),
	beta_reduce( Type1:Formula1, Type0:Formula0, Type:Formula ),
	nonvar( Formula ), % do we still need this?
	dbg( 'comp: eliminate at ' ), dbg( Pointer ), dbg( ' -> ' ), dbg( Formula ), dbgnl.


% CONJUNCTION (yes, all right, I just made this one up)
rule( conjunction,
      tree( Nodes, Pointer ), tree( [node( Pointer, [ty( CType ), fo( CFormula ), +linked | Labels2] ) | Nodes2], Pointer ) ) :-
	select( node( Pointer, Labels ), Nodes, Nodes2 ),
	\+ member( +linked, Labels ), % to stop this happening more than once
	select( ty( Type ), Labels, Labels1 ),
	select( fo( Formula ), Labels1, Labels2 ),
	dtr( 0, Pointer, D0 ),
	member( node( D0, _Labels0 ), Nodes ),
	linked_dtr( D0, DL ),
	member( node( DL, LabelsL ), Nodes ),
	member( ty( TypeL ), LabelsL ),
	member( fo( FormulaL ), LabelsL ),
	conjoin( TypeL:FormulaL, Type:Formula, CType:CFormula ),
	dbg( 'comp: conjoin at ' ), dbg( Pointer ), dbg( ' -> ' ), dbg( Formula ), dbgnl.


% COMPLETION (standard)
rule( completion,
      tree( Nodes, Pointer ), tree( Nodes, Mother ) ) :-
	member( node( Pointer, Labels ), Nodes ),
	member( ty( _Type ), Labels ),
	%member( fo( _Formula ), Labels ), % check formal definition
	(
	  dtr( _, Mother, Pointer );
	  linked_dtr( Mother, Pointer );
	  unfixed_dtr( Mother, Pointer )
	),
	dbg( 'comp: complete, up to ' ), dbg( Mother ), dbgnl.

% MERGE
rule( merge,
      tree( Nodes, Pointer ), tree( [node( Pointer, Labels ) | Nodes4], Pointer ) ) :-
	select( node( Pointer, FLabels ), Nodes, Nodes2 ),
	\+ member( ?fixed, FLabels ),
	(
	  \+ member( fo( FFormula ), FLabels ) % no formula
	;
	  member( fo( FFormula ), FLabels ), % pronoun
	  var( FFormula )
	),
	select( node( Node, ULabels ), Nodes2, Nodes3 ),
	select( ?fixed, ULabels, ULabels2 ),
	member( fo( _Formula ), ULabels2 ),
	find_node( Pointer, [/\ *,\/ *], Node ), % check no links in between, i.e. same subtree
%	subset( ULabels2, Labels ),
	union( FLabels, ULabels2, Labels ),
	compatible( Labels ),
	rename_nodes( Node, Pointer, Nodes3, Nodes4 ),
	dbg( 'comp: merge ' ), dbg( Node ), dbg( ' with ' ), dbg( Pointer ), dbgnl.

% RESOLUTION
rule( resolution,
      tree( Nodes, Pointer ), tree( Nodes, Pointer ) ) :-
	member( node( Pointer, Labels1 ), Nodes ),
	member( ty( Type ), Labels1 ),
	member( fo( Var ), Labels1 ),
 	var( Var ),
 	member( node( _Node, Labels2 ), Nodes ),
	member( ty( Type ), Labels2 ),
	member( fo( Var ), Labels2 ),
 	nonvar( Var ),
	union( Labels1, Labels2, Labels ),
	compatible( Labels ),
 	dbg( 'comp: resolve ' ), dbg( Pointer ), dbg( ' to ' ), dbg( Var ), dbgnl.
			   

rename_nodes( _Old, _New, [], [] ).
rename_nodes( Old, New, [node( O, Labels ) | OT], [node( N, Labels ) | NT] ) :-
	rename_node( Old, New, O, N ),
	rename_nodes( Old, New, OT, NT ).

rename_node( Old, _New, O, O ) :-
	\+ (
	     atom_concat( Mid, _End, O ),
	     atom_concat( _Start, Old, Mid )
	   ).
rename_node( Old, New, O, N ) :-
	atom_concat( Mid1, End, O ),
	atom_concat( Start, Old, Mid1 ),
	atom_concat( Start, New, Mid2 ),
	atom_concat( Mid2, End, N ).

union( L, R, U ) :-
	member( M, L ),
	select( M, R, R2 ),
	!,
	union( L, R2, U ).
union( L, R, U ) :-
	append( L, R, U ).

subset( [], _B ).
subset( S, B ) :-
	member( M, B ),
	select( M, S, S2 ),
	subset( S2, B ).

% "CORRECT" version - with type-checking, currently unusable as typechecking prevents
% conjoining cn & t formulae as is required in restrictive relatives
%
% % same type? just conjunction
% conjoin( Type:Formula1, Type:Formula2, Type:(Formula1 & Formula2) ).
% % not same type? recurse inside lambda-abstraction
% conjoin( Type1:Formula1, (ArgType>Type2):(Var^Formula2), (ArgType>Type):(Var^Formula) ) :-
% 	conjoin( Type1:Formula1, Type2:Formula2, Type:Formula ).

% "INCORRECT" version - no type-checking, just ensures conjunction done within all
% lambda-abstracted variables
%
% lambda-abstract? recurse inside lambda-abstraction
conjoin( Type1:Formula1, Type2:(Var^Formula2), Type:(Var^Formula) ) :-
	!,
	conjoin( Type1:Formula1, Type2:Formula2, Type:Formula ).
% run out of variables? just conjunction
conjoin( _Type1:Formula1, Type:Formula2, Type:(Formula1 & Formula2) ).
