%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% main module
%
% Matthew Purver, 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library( charsio ), [read_from_chars/2] ).

:- op( 900, fx, [?, /\, \/, if, then, else] ).
:- op( 800, xf, * ).
:- op( 1000, yfx, [&] ).

:- ensure_loaded( [lexical_templates, lexical_actions, computational_actions] ).
:- ensure_loaded( [lexicon_test, dstrees, lambda_calculus] ).
:- ensure_loaded( [parser, generator, io, printer, utils] ).

:- dynamic trees_to_print/1, format/1, debug/1, io/1.


% process_args/0
% Gets command-line arguments or sets defaults for display format etc.
process_args :-
	prolog_flag( argv, ArgList ),
	process_args( ArgList ),
	(
	  trees_to_print( _ )
	->
	  true
	;
	  % do we print all trees in forest, or only the complete ones? (all/some/complete)
	  assert( trees_to_print( complete ) )
	),
	(
	  format( _ )
	->
	  true
	;
	  % do we print in HTML format, or plain text? (txt/html)
	  assert( format( txt ) )
	),
	(
	  debug( _ )
	->
	  true
	;
	  % do we print debug messages (computational / lexical actions)? (on/off)
	  assert( debug( off ) )
	),
	(
	  io( _ )
	->
	  true
	;
	  % do we print to the terminal or to network sockets? (term/sock)
	  assert( io( term ) )
	).

% process_args/1 - process_args( +ArgList )
% Asserts any suitable command-line arguments
process_args( [] ).
process_args( [H | T] ) :-
	name( H, Chars ),
	read_from_chars( Chars, Term ),
	(
	  callable( Term )
	->
	  assert( Term )
	;
	  true
	),
	process_args( T ).


% message to halt parsing
end_message( [halt] ).
end_message( [exit] ).
% message to start generating
gen_message( [gen, Method, Number], [Method, Number] ).
% message to empty current forest
reset_message( [reset] ).


% ds/0
% Top-level - sets up initial forest and starts
ds :-
	process_args,
	startup( Socket ),
	root_forest( Trees ),
%	print_forest( Trees ),
	ds( Socket, Trees ),
	shutdown( Socket ).


% ds/2 - ds( ?Socket, +TreeList )
% Gets input from user and passes to ds/3
% Socket should be instantiated if io(sock) true
% TreeList should be the current parse forest
ds( Socket, Trees1 ) :-
	get_input( Socket, Words ),
	ds( Socket, Words, Trees1 ).

% ds/3 - ds( ?Socket, +WordList, +TreeList )
% Checks for system message, calls parser or generator as required
% Succeeds when: told to stop / told to generate / parse forest becomes empty
ds( _Socket, Words, _Trees ) :-
	end_message( Words ),
	!.

ds( Socket, Words, _Trees ) :-
	reset_message( Words ),
	!,
	root_forest( Trees ),
	ds( Socket, Trees ).

ds( _Socket, Words, Trees ) :-
	gen_message( Words, Method ),
	!,
	generate( Trees, Strings, Method ),
	newline, write( 'gen> ' ), print( Strings ), newline.

ds( Socket, Words, Trees1 ) :-
	statistics( runtime, _ ),
	parse( Trees1, Words, Trees2 ),
	statistics( runtime, [_,R] ),
	format( "Runtime ~w~n", [R] ),
	(
	  Trees2 = []
	->
	  print( 'No derivations' ), newline
	;
	  print_forest( Trees2 ),
	  current_output( Stream ),
	  close( Stream ),
	  ds( Socket, Trees2 )
	).
