%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic Syntax
%
% generator
%
% Matthew Purver, 2002
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library( lists ), [member/2, memberchk/2, append/3, remove_duplicates/2] ).
:- use_module( library( terms ), [variant/2] ).

% some predefined trees for testing
gen_trees( [
	    ( snore(Man), tree([node('0',[+pres,ty(t),fo(snore(Man))]),node('01',[ty(e>t),fo(A^snore(A))]),node('00',[+(male),ty(e),fo(Man)])],'0') ),
	    ( like(Man,Woman), tree([node('0',[+pres,ty(t),fo(like(Man,Woman))]),node('01',[ty(e>t),fo(A^like(A,Woman))]),node('010',[+(female),ty(e),fo(Woman)]),node('011',[ty(e>(e>t)),fo(B^C^like(C,B))]),node('00',[+(male),ty(e),fo(Man)])],'0') )
	   ] ).

% top-level: pull out the complete trees and generate them with generate/4
generate( Trees, String, Method ) :-
	findall( Tree, ( member( Tree, Trees ), complete( Tree ) ), GoalTrees ),
	root_forest( Root ),
	gen( GoalTrees, Root, Method, String ).


% base case
gen( [], _Root, _Method, [] ).
% depth-first, one solution
gen( [H | _T], Root, [d, one], String ) :-
	lexical_selection( H, W ),
	d_gen( H, W, Root, [], String ).
% depth-first, all solutions
gen( [H | _T], Root, [d, all], StringList ) :-
	lexical_selection( H, W ),
	findall( String, d_gen( H, W, Root, [], String ), StringList ).
% breadth-first
gen( [H | _T], Root, [b, Number], StringList ) :-
	lexical_selection( H, W ),
	b_gen( H, W, [sfp( [], Root )], StringList, Number ).


% main generation predicate. Simple or what?
main_gen( GoalTree, Word, Trees1, Trees2, String1, String2 ) :-
	word( Word, Action ),
	findall( Tree2, (
			  member( Tree1, Trees1 ),
			  exec( Tree1, Action, Tree1b ),
			  adjust( Tree1b, Tree2 ),
			  subsumes( GoalTree, Tree2 )
			), Trees2 ),
	Trees2 \= [],
	append( String1, [Word], String2 ). 
%	print( String2 ), nl.


% d_gen/2 - d_gen( +LogicalForm, ?String )
% just for testing
d_gen( Sem, String ) :-
	gen_trees( GT ),
	member( ( Sem, GoalTree ), GT ),
	root_forest( Trees ),
	d_gen( GoalTree, Trees, [], String ).

% d_gen/4 - gen( +CurrentTreeList, +GoalTree, +CurrentString, ?FinalString )
% depth-first search
d_gen( GoalTree, Trees, String, String ) :-
	member( Tree, Trees ),
	complete( Tree ),
%	variant( Tree, GoalTree ),
	subsumes( GoalTree, Tree ),
	newline, print( String ), newline.
d_gen( GoalTree, Trees1, String1, String ) :-
	main_gen( GoalTree, _Word, Trees1, Trees2, String1, String2 ),
	d_gen( GoalTree, Trees2, String2, String ).

% d_gen/5 - gen( +CurrentTreeList, +WordBag, +GoalTree, +CurrentString, ?FinalString )
% depth-first search with lexical pre-selection
d_gen( GoalTree, _WordBag, Trees, String, String ) :-
	member( Tree, Trees ),
	complete( Tree ),
%	variant( Tree, GoalTree ),
	subsumes( GoalTree, Tree ),
	newline, print( String ), newline.
d_gen( GoalTree, WordBag, Trees1, String1, String ) :-
	member( Word, WordBag ),
	main_gen( GoalTree, Word, Trees1, Trees2, String1, String2 ),
	d_gen( GoalTree, WordBag, Trees2, String2, String ).


% b_gen/2 - b_gen( +LogicalForm, ?StringList )
% just for testing
b_gen( Sem, StringList, Number ) :-
	gen_trees( GT ),
	member( ( Sem, GoalTree ), GT ),
	root_forest( Trees ),
	b_gen( GoalTree, [sfp( [], Trees )], StringList, Number ).


% b_gen/4 - breadth-first version
b_gen( _GoalTree, [], _StringList, _Number ).
b_gen( GoalTree, SFPList, StringList, Number ) :-
	findall( String, (
			   member( sfp( String, Trees ), SFPList ),
			   member( Tree, Trees ),
			   complete( Tree ),
			   variant( Tree, GoalTree ),
			   newline, print( String ), newline
			 ), StringList ),
	StringList \= [],
	Number = one.
b_gen( GoalTree, SFPList1, StringList, Number ) :-
	findall( sfp( String2, Trees2 ), (
					   member( sfp( String1, Trees1 ), SFPList1 ),
					   main_gen( GoalTree, _Word, Trees1, Trees2, String1, String2 )
					 ), SFPList2 ),
	b_gen( GoalTree, SFPList2, StringList, Number ).


% b_gen/5 - breadth-first version with lexical pre-selection
b_gen( _GoalTree, _WordBag, [], _StringList, _Number ).
b_gen( GoalTree, _WordBag, SFPList, StringList, Number ) :-
	findall( String, (
			   member( sfp( String, Trees ), SFPList ),
			   member( Tree, Trees ),
			   complete( Tree ),
			   variant( Tree, GoalTree ),
			   newline, print( String ), newline
			 ), StringList ),
	StringList \= [],
	Number = one.
b_gen( GoalTree, WordBag, SFPList1, StringList, Number ) :-
	findall( sfp( String2, Trees2 ), (
					   member( sfp( String1, Trees1 ), SFPList1 ),
					   member( Word, WordBag ),
					   main_gen( GoalTree, Word, Trees1, Trees2, String1, String2 )
					 ), SFPList2 ),
	b_gen( GoalTree, WordBag, SFPList2, StringList, Number ).


lexical_selection( tree( Nodes, _Pointer ), WordBag ) :-
	lexical_selection( Nodes, WordBag1 ),
	remove_duplicates( WordBag1, WordBag ),
	print( 'Bag: ' ), print( WordBag ), newline.

lexical_selection( [], [who, which, that] ).

lexical_selection( [node( _N, Labels ) | NTail], WordBag ) :-
	findall( Word, (
			 member( fo( Formula ), Labels ),
			 Formula \= e(_),
			 word( Word, Action ),
			 associated( Action, fo( Formula ) ),
			 \+ (
			      member( +Label, Labels ),
			      \+ associated( Action, +Label )
			    )
		       ), Words ),
	lexical_selection( NTail, WTail ),
	append( Words, WTail, WordBag ).


associated( put( Label1 ), Label2 ) :-
	variant( Label1, Label2 ).

associated( ( if _List1, then List2, else List3 ), Label ) :-
	associated( List2, Label );
	associated( List3, Label ).

associated( [Head | Tail], Label ) :-
	associated( Head, Label );
	associated( Tail, Label ).
